from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from inspect import iscoroutinefunction, signature
from random import random
from re import Pattern, RegexFlag, compile
from typing import Any, TypeVar

from pydantic import BaseModel

from chatushka._logger import logger
from chatushka._models import Events, Update
from chatushka._transport import TelegramBotAPI

Matcher = TypeVar("Matcher", bound="BaseMatcher")


class BaseMatcher(
    ABC,
):
    def __init__(
        self,
        action: Callable,
        chance_rate: float = 1.0,
        results_model: type[Any] | None = None,
    ) -> None:
        self._action: Callable = action
        self._chance_rate = chance_rate
        self._results_model = results_model

    @abstractmethod
    def _check(
        self,
        update: Update,
    ) -> list[str] | None:
        raise NotImplementedError

    async def __call__(
        self,
        api: TelegramBotAPI,
        update: Update,
    ) -> None:
        if (
            results := self._check(
                update=update,
            )
        ) is None:
            return
        logger.info(f"{self} matched with update_id={update.update_id} {results=}")
        await self._call_action(
            api=api,
            update=update,
            results=results,
        )

    def _get_chance(
        self,
    ) -> bool:
        return random() <= self._chance_rate

    def _make_results_model(
        self,
        results: list[str],
    ) -> Any:
        if self._results_model is None:
            return None
        if issubclass(self._results_model, BaseModel):
            params = {}
            for i, name in enumerate(self._results_model.model_fields):
                if len(results) < i:
                    break
                params[name] = results[i]
            return self._results_model.model_validate(params)
        return self._results_model(results)

    async def _call_action(
        self,
        api: TelegramBotAPI,
        update: Update,
        results: Any,
    ) -> None:
        if not self._get_chance():
            return
        kwargs = {}
        results_from_model = self._make_results_model(results)
        kwargs.update(
            {
                "api": api,
                "update": update,
                "message": update.message,
                "chat": update.message.chat if update.message else None,
                "user": update.message.user if update.message else None,
                "results": results_from_model or results,
            }
        )
        sig = signature(self._action)
        kwargs = {param: kwargs.get(param) for param in sig.parameters if param in kwargs}
        if update and update.message is not None and "message" in sig.parameters:
            kwargs["message"] = update.message
        if iscoroutinefunction(self._action):
            await self._action(**kwargs)
            return
        self._action(**kwargs)


class CommandMatcher(
    BaseMatcher,
    ABC,
):
    def __init__(
        self,
        *commands: str,
        action: Callable,
        prefixes: str | Sequence[str] = (),
        case_sensitive: bool = False,
        chance_rate: float = 1.0,
        results_model: type[Any] | None = None,
    ) -> None:
        super().__init__(
            action=action,
            chance_rate=chance_rate,
            results_model=results_model,
        )
        if case_sensitive:
            commands = tuple(command.upper() for command in commands)
        self._commands = commands
        self._case_sensitive = case_sensitive
        if prefixes:
            self.add_commands_prefixes(prefixes)

    def __repr__(
        self,
    ) -> str:
        return f"<{self.__class__.__name__}: {', '.join(self._commands)}>"

    def add_commands_prefixes(
        self,
        prefixes: Sequence[str],
    ) -> None:
        if not prefixes:
            return
        self._commands = tuple(f"{prefix}{command}" for command in self._commands for prefix in prefixes)

    def _make_args(
        self,
        text: str,
    ) -> list[str]:
        args = [arg for arg in text.split(" ") if arg]
        if len(args) == 1:
            return []
        return args[1:]

    def _check(
        self,
        update: Update,
    ) -> list[str] | None:
        if not update.message or not update.message.text:
            return None
        case_sensitive = self._case_sensitive or False
        for command in self._commands:
            if case_sensitive and update.message.text.upper().startswith(command.upper()):
                return self._make_args(update.message.text)
            if update.message.text.startswith(command):
                return self._make_args(update.message.text)
        return None


class RegExMatcher(
    BaseMatcher,
    ABC,
):
    def __init__(
        self,
        *patterns: str | Pattern,
        action: Callable,
        re_flags: int | RegexFlag = 0,
        chance_rate: float = 1.0,
        results_model: type[Any] | None = None,
    ) -> None:
        super().__init__(
            action=action,
            chance_rate=chance_rate,
            results_model=results_model,
        )
        self._patterns = [
            compile(pattern, flags=re_flags) if isinstance(pattern, str) else pattern for pattern in patterns
        ]

    def __repr__(
        self,
    ) -> str:
        return f"<{self.__class__.__name__}: {', '.join([entry.pattern for entry in self._patterns])}>"

    def _check(
        self,
        update: Update,
    ) -> list[str] | None:
        if not update.message or not update.message.text:
            return None
        for pattern in self._patterns:
            if result := pattern.findall(update.message.text):
                return result
        return None


class EventMatcher(
    BaseMatcher,
    ABC,
):
    def __init__(
        self,
        *events: Events,
        action: Callable,
        chance_rate: float = 1.0,
        results_model: type[Any] | None = None,
    ) -> None:
        super().__init__(
            action=action,
            chance_rate=chance_rate,
            results_model=results_model,
        )
        self._events = events

    def __repr__(
        self,
    ) -> str:
        return f"<{self.__class__.__name__}: {', '.join(self._events)}>"

    def _check(
        self,
        update: Update,
    ) -> list[str] | None:
        results = []
        if update.message and update.message.text and "on_message" in self._events:
            results.append("on_message")
        if update.message and update.message.new_chat_members and "on_new_chat_members" in self._events:
            results.append("on_new_chat_members")
        if update.message and update.message.new_chat_members and "on_left_chat_member" in self._events:
            results.append("on_left_chat_member")
        if not results:
            return None
        return results



from datetime import datetime
from types import TracebackType
from typing import Any, Literal

from httpx import AsyncClient, Response, TimeoutException

from chatushka._constants import HTTP_REGULAR_TIMEOUT
from chatushka._errors import ChatushkaResponseError
from chatushka._logger import logger
from chatushka._models import (
    ChatMemberAdministrator,
    ChatMemberOwner,
    ChatMemberStatuses,
    ChatPermissions,
    Message,
    Update,
)
from chatushka._sentry import report_exc


async def _raise_on_api_error_response_event_hook(
    response: Response,
) -> None:
    await response.aread()
    if not response.is_success:
        raise ChatushkaResponseError(
            response=response,
        )
    data: dict[str, Any] = response.json()
    if data.get("ok", False) is False or data.get("result") is None:
        raise ChatushkaResponseError(
            response=response,
        )


class TelegramBotAPI:
    _offsets: dict[str, int] = {}  # noqa

    def __init__(
        self,
        token: str,
        timeout: int = HTTP_REGULAR_TIMEOUT,
    ) -> None:
        self._token = token
        self._client = AsyncClient(
            base_url=f"https://api.telegram.org/bot{token}",
            event_hooks={  # type: ignore
                "response": [
                    _raise_on_api_error_response_event_hook,
                ]
            },
            timeout=timeout,
        )
        self._timeout = timeout

    def __repr__(
        self,
    ) -> str:
        return f"<{self.__class__.__name__}>"

    async def _api_request(
        self,
        api_method: str,
        **kwargs: Any,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        logger.debug(f"{self} >>> {api_method} >>> {kwargs}")
        response = await self._client.post(
            url=api_method,
            json=kwargs,
        )
        data = response.json()["result"]
        logger.debug(f"{self} <<< {api_method} <<< {response.text}")
        return data

    async def __aenter__(
        self,
    ) -> "TelegramBotAPI":
        await self._client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        return await self._client.__aexit__(
            exc_type=exc_type,
            exc_value=exc_value,
            traceback=traceback,
        )

    async def get_updates(
        self,
        offset: int | None,
    ) -> tuple[list[Update], int | None]:
        params = {} if not offset else {"offset": offset}
        if self._timeout:
            params["timeout"] = self._timeout
        try:
            response = await self._api_request(
                api_method="getUpdates",
                **params,
            )
        except TimeoutException:
            return [], offset
        except Exception as exc:
            report_exc(exc)
            logger.error(f"{self} !!! {exc.__class__.__name__} ({exc!s})")
            return [], offset
        results = [Update.model_validate(entry) for entry in response]
        if results:
            offset = results[-1].update_id + 1
        return results, offset

    async def send_message(
        self,
        chat_id: int,
        text: str,
        reply_to_message_id: int | None = None,
        parse_mode: Literal["html", "markdown"] = "html",
        disable_web_page_preview: bool = False,
    ) -> Message:
        result = await self._api_request(
            api_method="sendMessage",
            chat_id=chat_id,
            text=text,
            reply_to_message_id=reply_to_message_id,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview,
        )
        return Message.model_validate(result)

    async def get_chat_administrators(
        self,
        chat_id: int,
    ) -> list[ChatMemberAdministrator | ChatMemberOwner]:
        results = await self._api_request(
            "getChatAdministrators",
            chat_id=chat_id,
        )
        admins: list[ChatMemberAdministrator | ChatMemberOwner] = []
        for result in results:
            status = result["status"]  # type: ignore
            if status == ChatMemberStatuses.CREATOR:
                admins.append(ChatMemberOwner.model_validate(result))
            if status == ChatMemberStatuses.ADMINISTRATOR:
                admins.append(ChatMemberAdministrator.model_validate(result))
        return admins

    async def restrict_chat_member(
        self,
        chat_id: int,
        user_id: int,
        permissions: ChatPermissions,
        until_date: datetime,
    ) -> bool:
        return await self._api_request(  # type: ignore
            "restrictChatMember",
            chat_id=chat_id,
            user_id=user_id,
            permissions=permissions.json(),
            until_date=int(until_date.timestamp()),
        )



from collections.abc import Callable
from re import RegexFlag
from typing import Any, final

from chatushka._bot import MatchersBot
from chatushka._matchers import CommandMatcher, EventMatcher, RegExMatcher
from chatushka._models import Events


@final
class Chatushka(
    MatchersBot,
):
    def cmd(
        self,
        *commands: str,
        case_sensitive: bool = False,
        chance_rate: float = 1.0,
        results_model: type[Any] | None = None,
    ) -> Callable:
        def _wrapper(
            func,
        ) -> None:
            self.add(
                CommandMatcher(
                    *commands,
                    action=func,
                    case_sensitive=case_sensitive,
                    chance_rate=chance_rate,
                    results_model=results_model,
                )
            )

        return _wrapper

    def regex(
        self,
        *patterns: str,
        re_flags: int | RegexFlag = 0,
        chance_rate: float = 1.0,
        results_model: type[Any] | None = None,
    ) -> Callable:
        def _wrapper(
            func,
        ) -> None:
            self.add(
                RegExMatcher(
                    *patterns,
                    action=func,
                    re_flags=re_flags,
                    chance_rate=chance_rate,
                    results_model=results_model,
                )
            )

        return _wrapper

    def event(
        self,
        *events: Events,
        chance_rate: float = 1.0,
        results_model: type[Any] | None = None,
    ) -> Callable:
        def _wrapper(
            func,
        ) -> None:
            self.add(
                EventMatcher(
                    *events,
                    action=func,
                    chance_rate=chance_rate,
                    results_model=results_model,
                )
            )

        return _wrapper



from contextlib import suppress

from httpx import RequestNotRead

from chatushka._errors import ChatushkaResponseError

try:
    import sentry_sdk  # type: ignore
except ImportError:
    sentry_sdk = None  # type: ignore


def report_exc(
    exc: Exception,
) -> None:
    with sentry_sdk.push_scope() as scope:
        if isinstance(exc, ChatushkaResponseError):
            scope.set_extra(
                "response",
                {
                    "status": exc.response.status_code,
                    "data": exc.response.json(),
                },
            )
            with suppress(UnicodeDecodeError, RequestNotRead):
                scope.set_extra(
                    "request",
                    exc.response.request.content.decode(),
                )
            sentry_sdk.capture_exception(exc)



from chatushka._chatushka import Chatushka
from chatushka._errors import ChatushkaError, ChatushkaResponseError
from chatushka._matchers import BaseMatcher, CommandMatcher, EventMatcher, RegExMatcher
from chatushka._models import Chat, ChatPermissions, Events, Message, Update, User
from chatushka._transport import TelegramBotAPI

__version__ = "0.11rc15"
__all__ = [
    "Chatushka",
    # errors
    "ChatushkaError",
    "ChatushkaResponseError",
    # matchers
    "BaseMatcher",
    "CommandMatcher",
    "EventMatcher",
    "RegExMatcher",
    # models
    "Events",
    "Chat",
    "ChatPermissions",
    "Message",
    "Update",
    "User",
    # transport
    "TelegramBotAPI",
]



from datetime import datetime
from enum import Enum
from typing import Literal, Optional

from pydantic import AliasChoices, BaseModel, ConfigDict, Field


class ChatType(
    str,
    Enum,
):
    GROUP = "group"
    SUPERGROUP = "supergroup"
    PRIVATE = "private"


class ChatMemberStatuses(
    str,
    Enum,
):
    MEMBER = "member"
    CREATOR = "creator"
    ADMINISTRATOR = "administrator"
    LEFT = "left"
    KICKED = "kicked"
    RESTRICTED = "restricted"


class User(
    BaseModel,
):
    id: int
    is_bot: bool
    first_name: str | None = None
    last_name: str | None = None
    can_join_groups: bool | None = None
    can_read_all_group_messages: bool | None = None

    @property
    def readable_name(self) -> str:
        return f"{self.first_name}{' '+self.last_name if self.last_name else ''}"


class ChatMemberBase(
    BaseModel,
):
    status: ChatMemberStatuses
    user: User
    is_anonymous: bool | None = None
    custom_title: str | None = None


class ChatMemberOwner(
    ChatMemberBase,
): ...


class ChatMemberAdministrator(
    ChatMemberBase,
):
    can_be_edited: bool | None = None
    can_manage_chat: bool | None = None
    can_delete_messages: bool | None = None
    can_manage_voice_chats: bool | None = None
    can_restrict_members: bool | None = None
    can_promote_members: bool | None = None
    can_change_info: bool | None = None
    can_invite_users: bool | None = None
    can_post_messages: bool | None = None
    can_edit_messages: bool | None = None
    can_pin_messages: bool | None = None


class Chat(
    BaseModel,
):
    id: int
    type: ChatType
    title: str | None = None


class NewChatMember(
    ChatMemberAdministrator,
): ...


class OldChatMember(
    ChatMemberAdministrator,
): ...


class Message(
    BaseModel,
):
    message_id: int
    user: User = Field(
        validation_alias=AliasChoices("user", "from"),
    )
    chat: Chat
    text: str | None = None
    reply_to_message: Optional["Message"] = None
    new_chat_members: list[User] = Field(default_factory=list)


Message.update_forward_refs()


class MyChatMember(
    BaseModel,
):
    chat: Chat
    user: User = Field(..., alias="from")
    date: datetime
    old_chat_member: OldChatMember
    new_chat_member: NewChatMember


class Update(
    BaseModel,
):
    model_config = ConfigDict(
        populate_by_name=True,
    )
    update_id: int
    message: Message | None = None
    my_chat_member: MyChatMember | None = None


class ChatPermissions(
    BaseModel,
):
    can_send_messages: bool
    can_send_media_messages: bool
    can_send_polls: bool
    can_send_other_messages: bool


Events = Literal[
    "on_message",
    "on_new_chat_members",
    "on_left_chat_member",
]



from httpx import Response


class ChatushkaError(
    Exception,
):
    pass


class ChatushkaResponseError(
    ChatushkaError,
):
    def __init__(
        self,
        response: Response,
    ) -> None:
        self.response = response
        super().__init__(f'Telegram BOT API response error:\n  {response.text}"')



import signal
from asyncio import ensure_future, gather, get_event_loop
from collections.abc import AsyncGenerator, MutableMapping, Sequence
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from datetime import timezone, tzinfo
from traceback import print_exception

import aiocron  # type: ignore

from chatushka._constants import HTTP_POOLING_TIMEOUT, HTTP_REGULAR_TIMEOUT
from chatushka._errors import ChatushkaResponseError
from chatushka._logger import logger
from chatushka._matchers import CommandMatcher, Matcher, TelegramBotAPI
from chatushka._sentry import report_exc


@asynccontextmanager
async def _default_lifespan(
    _: "MatchersBot",
) -> AsyncGenerator[None, None]:
    yield


class MatchersBot:
    def __init__(
        self,
        *,
        token: str,
        cmd_prefixes: str | Sequence[str] = (),
        lifespan: AbstractAsyncContextManager | None = None,
    ) -> None:
        self._state: MutableMapping = {}
        self._lifespan = lifespan or _default_lifespan
        self._token = token
        if isinstance(cmd_prefixes, str):
            cmd_prefixes = [cmd_prefixes]
        self._cmd_prefixes = cmd_prefixes
        self._matchers: list[Matcher] = []  # type: ignore
        self._schedulers: list[aiocron.Cron] = []
        self._closed: bool = True

    def __repr__(
        self,
    ) -> str:
        return f"<{self.__class__.__name__}: {len(self._matchers)} matchers, {len(self._schedulers)} schedulers>"

    def add(
        self,
        matcher: Matcher,
    ) -> None:
        if isinstance(matcher, CommandMatcher):
            matcher.add_commands_prefixes(
                self._cmd_prefixes,
            )
        logger.info(f"{self} + {matcher}")
        self._matchers.append(matcher)

    async def _call_scheduled(
        self,
        func,
    ):
        logger.info(f"{self} ::: scheduled call of {func.__name__}")
        async with self._make_api_client() as client:
            await func(
                api=client,
            )

    def schedule(
        self,
        cron: str,
        tz: tzinfo = timezone.utc,
    ):
        def _wrapper(
            func,
        ):
            job = aiocron.Cron(
                spec=cron,
                func=self._call_scheduled,
                start=False,
                tz=tz,
                args=(func,),
            )
            logger.info(f"{self} + {job}")
            self._schedulers.append(job)

        return _wrapper

    async def _check_updates(
        self,
        api: TelegramBotAPI,
        offset: int | None,
    ) -> int | None:
        try:
            updates, offset = await api.get_updates(offset)
        except (Exception, ChatushkaResponseError) as exc:
            report_exc(exc)
            return offset
        if not updates:
            return offset
        logger.debug(f"{self} <<< {len(updates)} updates from {offset=}")
        results = await gather(
            *[
                matcher(  # type: ignore
                    api=api,
                    update=update,
                )
                for update in updates
                for matcher in self._matchers
            ],
            return_exceptions=True,
        )
        for result in results:
            if isinstance(result, Exception):
                print_exception(result)
                report_exc(result)
        return offset

    async def _loop(
        self,
    ) -> None:
        offset: int | None = None
        while not self._closed:
            async with TelegramBotAPI(
                token=self._token,
                timeout=HTTP_POOLING_TIMEOUT,
            ) as api:
                offset = await self._check_updates(
                    api=api,
                    offset=offset,
                )

    def _make_api_client(
        self,
        timeout: int = HTTP_REGULAR_TIMEOUT,
    ) -> TelegramBotAPI:
        return TelegramBotAPI(
            token=self._token,
            timeout=timeout,
        )

    async def _stop(
        self,
    ) -> None:
        self._closed = True
        if self._schedulers:
            logger.info(f"{self} (っ◔◡◔)っ stop schedulers")
        for scheduler in self._schedulers:
            scheduler.stop()

    async def run(
        self,
    ) -> None:
        loop = get_event_loop()
        for sig in (signal.SIGINT, signal.SIGTERM):
            try:
                loop.add_signal_handler(sig, callback=lambda: ensure_future(self._stop()))
            except NotImplementedError:
                break
        if self._schedulers:
            logger.info(f"{self} (っ◔◡◔)っ start schedulers")
        for scheduler in self._schedulers:
            scheduler.start()
        logger.info(f"{self} (っ◔◡◔)っ start polling")
        self._closed = False
        async with self._lifespan(self):  # type: ignore
            await self._loop()
